/*
 The search for section 6.2 of https://www.mathi.uni-heidelberg.de/~yves/Papers/radical.html

 Please also look at the ReadMe:
 https://www.mathi.uni-heidelberg.de/~yves/Papers/radical_data/aux/ReadMe.pdf
*/

#include <stdlib.h>
#include <fstream>
#include <vector>
#include <array>
#include <bitset>

using namespace std;

#include "DHO.h"


template<size_t rank,  size_t dimU1, size_t dimU> //dimU1= dim(U(T1)). dimU= dim(U(S)).
class SearchHindSupDHO{
public:
	static_assert(rank>2,"Error: rank<3"); // for hat a suDHO exists.
	static_assert(dimU>dimU1,"Error: dimU<=dimU1");
	static_assert(dimU<=dimU1+rank,"Error: dimU>dimU1+rank");
	
	static const size_t n_r=(1<<rank);	// 2^rank
	static const size_t n_U=(1<<dimU);	// 2^dim
	static const size_t n_rh=n_r/2;		// 2^{rank-1}
	static const size_t dim_a = dimU-dimU1; // Dimension of T_2 "outside" of U(T_1)
	static const size_t dim_i = rank-dim_a; // Dimension of T_2 \cap U(T_1)
	
	typedef DHO_T<rank-1,dimU1> subDHO;
	typedef DHO_T<rank,dimU> DHO;
	typedef vec2T<dimU> U_T; // ambient space
	typedef vs2T<U_T> VR_T; // vector space in ambient space
	typedef vec2T<dimU1> sU_T; // ambient space of T_1
	typedef vs2T<sU_T> sVR_T; // vector spaces in the ambient space of T_1
	
	
	static constexpr const U_T e=U_T::e(dimU-1); // the hyperplane will be e^\perp (indices run from 0...dimU-1)
	subDHO T1;
	typename subDHO::inDHO_T in_T1;
	array<VR_T,n_rh+1> S0; // S(T_1) and Z
	VR_T Z0;
	array<sVR_T,dim_i+1> W; // W[dim_i] = Z \cap \cal C. The W[i] with smaller i are subspaces of W used for its construction
	array<VR_T,n_rh> S1;
	
	array< array<vs2T<U_T>, n_rh>, n_rh> TestP; // for testP()
	
	// counters for the results
	long andho {0};
	long nr_Type_H{0};
	long nr_Type_D{0};
	long nr_other{0};
	
	DHO S;
	
	
	
	SearchHindSupDHO( ifstream& fi)
	{
		//////////////////////////////////////////////////
		// (Step 0) initialization
		//////////////////////////////////////////////////
		
		// reads and initialize the subDHO T1
		fi>>T1;
		T1.init();
		//cout<<T1<<endl;
		in_T1=T1.get_inDHO(); // incidence vector of the points in T1.
		
		// Writes the T1-part to S0
		for (size_t i=0; i<n_rh;i ++){
			S0[i].reset();
			for (size_t j=0; j<rank-1;j ++)
			{
				S0[i].add(U_T::embed(T1[i].base(j)));
			}}
		
		// gernerates data struktur for testP()
		// TestP[i][j] = \langle T1[i],T1[j] \rangle
		{
			for (size_t i=0; i<n_rh; i++)
				for (size_t j=0; j<n_rh; j++)
				{
					TestP[i][j]=VR_T::embed(T1[i]);
					TestP[i][j].add(VR_T::embed(T1[j]));
				}
		}
		
		W[0].reset();
		
		//////////////////////////////////////////////////
		// Launch the search
		//////////////////////////////////////////////////
		rec_search_W(0,1); // recursively generates the dim_i-spaces "W" = Z \cap {\cal C}.
	}
	
	
	//////////////////////////////////////////////////
	// (Step 1) Generation of W and Z
	//////////////////////////////////////////////////
	
	void rec_search_W(const size_t ti, const size_t li) // recursively generates the dim_i-spaces "W" = Z \cap {\cal C}.
	// For all obtained Z the starter set S0 is generated and the search for siutable permutaions \pi is launched.
	{
		if (ti == dim_i){ // a suiteable dim_i-Raum W is found.
			{
				init1();  // Generates Z0=<V_0,W> and S0=<Z0,e>
				initpi(); // Generates the permutaions pi for the startersets and tests them.
			}}
		else
		{
			for (size_t iw=li; iw<n_U; iw++)
				if (! in_T1[iw]) // iw in \cal C
				{
					W[ti+1]=W[ti];
					W[ti+1].add(sU_T(iw));
					if (sU_T(iw)==W[ti+1].base(ti)) // For that the vector space W in \cal C  gets genterated only once.
						if(testW(ti)) // Test if the whole space W[ti+1] is in \cal C.
							rec_search_W(ti+1, iw+1);
				}
		}
	}
	
	bool testW(const size_t ti) const // Test if the whole space W[ti+1] is in \cal C.
	{
		sU_T v=W[ti+1].base(ti);
		for (size_t i=1; i<(size_t(1)<<W[ti].dim()); i++) {
			if (in_T1[size_t(W[ti].pkt(i)+v)] )
				return false;
		}
		return true;
	}
	
	
	
	void init1() // Generates Z0=<V_0,W> and S0=S[T] \cup <Z0,e>
	{
		// cout<<"-----------------------\nW in C: "<< W[dim_i];
		Z0.reset();
		for (size_t i=dimU1; i<dimU-1;i ++)
			Z0.add(U_T::e(i));
		
		Z0.add(VR_T::embed(W[dim_i]));
		//cout<<"Z0 "<<Z0<<endl;
		
		S0[n_rh]=Z0;
		S0[n_rh].add(e);
		S[n_rh]=S0[n_rh];
		S1[0]=S0[n_rh];
		//cout<<"Z "<<S0[n_rh]<<endl;
	}
	
	//////////////////////////////////////////////////
	// (Step 1) Generation of \pi
	//////////////////////////////////////////////////
	/*
	 We generate all permutaions pi recursively in a rather simple way:
	 For pi[ti] we loop over all i and and set pi[ti]=i if i \not= p[j], j<ti.
	 "inpi" is the incidence vector of the p[j], j<ti.
	 
	 Only such pemutaions are generated that satisy conditions given in the paper.
	 */
	
	bitset<n_rh> inpi;
	//typedef permutationT<n_rh> pT;
	std::array<size_t,n_rh> pi;
	
	void initpi() // Generates the permutaions pi for the startersets and tests them.
	{
		// some inits
		inpi.reset();
		pi[0]=0; // e
		inpi.set(0);
		
		// launch the search
		rec_search_Pi(1);
	}
	
	
	void rec_search_Pi(const int ti) // the actual recursive generation of pi for Z0
	{
		if (ti==n_rh) {
			if (genE()) // generates the \cal E_i. Returns false if one \cal E_i is empty.
			{
				genS1(1);// generates the resultion DHO;
			}
		}
		else
		{
			for (size_t i=1; i<n_rh;i ++) // due to tranitivity we can assume pi(0)=0.
				if (! inpi[i]) {
					pi[ti]=i;
					if(testP(ti)) // tests if pi has the required properties
					{
						inpi.set(i);
						rec_search_Pi(ti+1);
						inpi.reset(i);
					}
				}
		}
	}
	
	bool testP(const size_t ti) const // tests if pi has the required properties
	// i.e. checks if X^\pi_{ti} + X^\pi_{j}  \in \langle T1[ti],T1[j] \rangle
	{
		for (size_t j=0; j<ti;j ++)
			if(TestP[ti][j].is_element(Z0.pkt(pi[ti])+Z0.pkt(pi[j])) )
				return false;
		return true;
	}
	
	//////////////////////////////////////////////////
	// (Step 2) Generation of E_i
	//////////////////////////////////////////////////
	
	array<vector<VR_T>,n_rh-1> E;
	
	// Incidence vektors for test of disjunctness of the  E[i].
	// S0[ti].pkt(i)+e+Z0.pkt(pi[ti]) <-> bit ti*n_rh+i in incidence vector
	typedef bitset<n_rh*n_rh> ind12T;
	ind12T in, tin;
	array<vector<ind12T>,n_rh-1> iE;
	
	bool genE() // generates the {\cal E_i}. Returns false if one {\cal E_i} is empty.
	{
		tin.reset();
		for (size_t i=0; i<n_rh;i ++)
		{
			// writes the S0-part to the result DHO S
			S[i]=S0[i];
			S[i].add(e+Z0.pkt(pi[i]));
			
			tin.set(i*n_rh); // used in genS1()
		}
		
		// .add() may alter the bases of S[j] and thus .pkt(i) may result in an other vektor than before .add().
		// Therfore below "S0[j].pkt(i)+e+Z0.pkt(pi[j])" can not be replaced by "S[j].pkt(n_rh + i)".
		
		// Generation of  E_i
		for (size_t i=1; i<n_rh;i++) {
			// inits
			E[i-1].resize(0);
			iE[i-1].resize(0);
			VR_T V;
			in.reset();
			V.add(S0[0].pkt(i)+e+Z0.pkt(pi[0])); // E(i-1) are the candidate of S which intersect X0 in S[0].pkt(i)+e+Z0.pkt(pi[0])) .
			in.set(i);
			
			// recursive search for vectorspaces V in E[i-1]
			rec_search_V(i-1,1,V);
			if (E[i-1].size()==0) // if E_i is empty
			{
				return false;
			}
		}
		return true;
	}
	
	void rec_search_V(const int ind,const int ti, const VR_T& V)
	{
		if (ti==n_rh) // a candidate V is found
		{
			// cout<<" rec_search_V "<<V<<endl;
			if(test_intersection(V)) // tests that dim(V\cap X)=1 for all X \in S_0
			{
				E[ind].push_back(V);
				iE[ind].push_back(in);
			}
		}
		else
		{
			for (size_t i=1; i<n_rh;i++) // for i=0 we get the intersection with Z
			{
				VR_T vL{V};
				U_T v= S0[ti].pkt(i)+e+Z0.pkt(pi[ti]);
				vL.add(v);
				if (vL.dim()<=rank)
				{
					const size_t c=ti*n_rh + i;
					in.set(c);
					rec_search_V(ind,ti+1,vL);
					in.reset(c);
				}
			}
		}
	}
	
	bool test_intersection(const VR_T& V) const
	{
		for (size_t j=0; j<=n_rh; j++)
			if (1!=dim_meet(V,S[j]))
				return false;
		return true;
	}
	
	
	//////////////////////////////////////////////////
	// (Step 2) Generation of S1
	//////////////////////////////////////////////////
	
	void genS1(const size_t ti)
	{
		if (ti==n_rh) {
			mkDHO();
		}
		else
		{
			for (size_t i=0; i<E[ti-1].size();i++) {
				S1[ti]=E[ti-1][i];
				if ((tin & iE[ti-1][i]).none()) // disjointnes
					if (testT(ti))
					{
						tin|=iE[ti-1][i];
						genS1(ti+1);
						tin&=~iE[ti-1][i];
					}
			}
		}
	}
	
	bool testT(const size_t ti) const
	{
		
		// any 2 S1 intersect 1-dimensional
		for (size_t j=1; j<ti; j++)
			if (1!=dim_meet(S1[j],S1[ti])) // i=0 teste at genE()
				return false;
		// any 3 S1 intersect trivial
		for (size_t j=1; j<ti; j++)
			for (size_t i=0; i<j; i++)
				if (meet(S1[i],S1[ti])==meet(S1[j],S1[ti])) {
					return false;
				}
		return true;
	}
	
	
	//////////////////////////////////////////////////
	// (Step 2) Generation of the result (the DHO S)
	//////////////////////////////////////////////////
	
	bool mkDHO()
	{
		for (size_t i=1; i<n_rh; i++)
			S[i+n_rh]=S1[i];
		S.resetPU();
		
		if(S.DHO_test()){ // verbose
			cout<<"******* result "<<andho+1<<" ******* is DHO \n";
			S.init_P();
			S.init_U();
			if(S.is_type_H())
			{
				nr_Type_H++;
			}
			else if(S.is_type_D())
			{
				nr_Type_D++;
			}
			else
			{
				nr_other++;
			}
			andho++;
			return true;
		}
		else
		{
			cout<<"\n\nError! " <<andho+1<< " is no DHO "<<endl;
			exit(0);
		}
		return false;
	}
	
	
};

int main()
{
	ifstream fi("H3Data.txt");
	
	// usage: SearchHindSupDHO<rank, dim(SubDHO), dim(DHO)> SSDHO(fi);
	
	// Only uncomment one of the two following lines.
	// SearchHindSupDHO<4,6,9> SSDHO(fi); // launches the serch for dim(S)=9
	SearchHindSupDHO<4,6,10> SSDHO(fi); // launches the serch for dim(S)=10
	
	cout<<"total number of DHO     = "<<SSDHO.andho<<endl;
	cout<<"number of DHO of Type H = "<<SSDHO.nr_Type_H <<endl;
	cout<<"number of DHO of Type D = "<<SSDHO.nr_Type_D <<endl;
	cout<<"number of other DHO     = "<<SSDHO.nr_other <<endl;

	return 0;
}


