/*
 The search for section 6.2 of https://www.mathi.uni-heidelberg.de/~yves/Papers/radical.html
 
 Please also look at the ReadMe:
 https://www.mathi.uni-heidelberg.de/~yves/Papers/radical_data/aux/ReadMe.pdf
 */

#ifndef _vs2T_h
#define _vs2T_h

#include <iostream>
#include <vector>

#include "vec2.h"


template<class T>
std::ostream& operator<<(std::ostream& s, const std::vector<T>& v)
{
	for (unsigned int i=0; i<v.size(); i++)
		s<<v[i]<<" ";
	
	s<<"\n";
	return s;
}

// A class for binary vector spaces
template<typename uvt> // uvt type of the vector of the ambient space
class vs2T{
	
public:
	typedef uvt U_T; // vectors of the vector space
	static const size_t dimU {U_T::dimU}; // the dimension of the ambient space
private:
	typedef std::bitset<(1<<dimU)> I_T; // incidence vektor for subset in the ambient space
	std::vector<U_T> vr; // basis of the current vector space
	std::vector<size_t> vr_i; // vr_i[i] the position of the first 1 in vr[i]
	
public:
	vs2T()=default;
	~vs2T()=default;
	vs2T(const vs2T&)=default;
	vs2T(vs2T&&)=default;
	constexpr vs2T& operator=(const vs2T&)=default;
	constexpr vs2T& operator=(vs2T&&)=default;
	
	
	constexpr friend bool operator==(const vs2T& U,const vs2T& V)
	{
		if (U.dim()!=V.dim())
			return false;
		return U.dim()==span(U,V).dim();;
	}
	
	constexpr friend bool operator!=(const vs2T& a,const vs2T& b)
	{
		return !(a==b);
	}
	
	void reset(const size_t s=0)
	{
		vr.resize(s);
		vr_i.resize(s);
	}
	
	friend std::ostream& operator<<(std::ostream& s, const vs2T& V)
	{
		s<<V.vr;
		return s;
	}
	
	template<typename uvt2>
	static constexpr vs2T  embed(const vs2T<uvt2> v, const size_t offs=0)
	{ // embed an vector space v of an smaller ambient sapce in the current space.
		vs2T r;
		for(size_t i=0;i<v.dim();i++)
			r.add(vs2T::U_T::embed(v.base(i),offs));
		
		return r;
	}
	
	bool is_element(const U_T v) const
	{
		U_T lv {v};
		for (size_t i=0; i<vr.size(); i++)
			if(bool(lv[vr_i[i]]))
				lv-=vr[i];
		
		return lv.is_zero();
	}
	
	
	size_t index(const U_T v) const
	{ // the i such that pkt(i)=v
		if (v.is_zero())
			return 0;
		
		U_T lv {v};
		size_t ind{0};
		for (size_t i=0; i<vr.size(); i++)
			if(bool(lv[vr_i[i]]))
			{
				ind+= (1<<i);
				lv-=vr[i];
			}
		if ((ind==0)||(! lv.is_zero()))
		{
			std::cerr<<"error index "<<v<<std::endl;
			exit(1);
		}
		return ind;
	}
	
	void add(const U_T& v)
	{	// *this= \langle this,v \rangle
		// basis is put in standard form
		if (vr.size()==dimU) {
			return;
		}
		
		U_T lv {v};
		for (size_t i=0; i<vr.size(); i++)
			if(lv[vr_i[i]])
				lv+=vr[i];
		
		if(!lv.is_zero())
		{
			vr_i.push_back(lv.first());
			for (size_t i=0; i<vr.size(); i++)
				if(bool(vr[i][vr_i.back()]))
					vr[i]-=lv;
			
			vr.push_back(lv);
		}
	}
	
	void add(const vs2T& vv)
	{ // *this= \langle this,vv \rangle
		for (size_t i=0; i<vv.vr.size(); i++)
			add(vv.vr[i]);
	}
	
	size_t dim() const
	{
		return vr.size();
	}
	
	U_T base(const size_t i) const
	{
		return vr[i];
	}
	
	U_T pkt(const size_t j) const // For x =vec2T(j) x*S_j=pkt(j)
	{
		U_T p {0};
		for (size_t i=0; i<size(vr); i++)
			if((j>>i)&1)
				p+=vr[i];
		
		return p;
	}
	
	vs2T friend span(const vs2T& A,  const vs2T& B)
	{// A + B
		vs2T V {A};
		V.add(B);
		return V;
	}
	
	size_t friend dim_meet(const vs2T& A,  const vs2T& B)
	{ // dim(A meet B)
		return A.dim() + B.dim() - span(A,B).dim();
	}
	
	int friend span_and_meet(const vs2T& A,  const vs2T& B, vs2T& RV,  vs2T& RS)
	{ // RV= A+B, RS=A meet B
		vs2T V {A};
		vs2T T {B};
		vs2T VA {A}; // (will be the intersection)
		vs2T TA {B}; // (will be the intersection)
		vs2T S;
		int dim_S=0;
		
		// reduces the matrix A|B to a standart basis.
		// the zero rows after transformation correspond to the intersection
		// So finally
		// V|T, without the zero-rows will be the span.
		// It is T_* = Sum a_i A_i + b_i B_i
		// If T_* = 0 then Sum a_i A_i is an element of the intersection.
		
		for (size_t j=0; j<TA.vr.size(); j++)
			TA.vr[j]=U_T::zero;
		
		for (size_t j=0; j<T.vr.size(); j++)
		{
			if(T.vr[j].is_zero())
				continue;
			for (size_t i=0; i<V.vr.size(); i++)
				if(bool(T.vr[j][V.vr_i[i]]))
				{
					T.vr[j]+=V.vr[i];
					TA.vr[j]+=VA.vr[i];
				}
			T.vr_i[j]=T.vr[j].first();
			if(!T.vr[j].is_zero())
			{
				for (size_t i=0; i<V.vr.size(); i++)
					if(bool(V.vr[i][T.vr_i[j]]))
					{
						V.vr[i]+=T.vr[j];
						VA.vr[i]+=TA.vr[j];
					}
				for (size_t i=0; i<T.vr.size(); i++)
					if(i!=j)
						if(bool(T.vr[i][T.vr_i[j]]))
						{
							T.vr[i]+=T.vr[j];
							TA.vr[i]+=TA.vr[j];
						}
			}
			else
			{
				dim_S++;
			}
		}
		
		for (size_t j=0; j<T.vr.size(); j++)
			if (T.vr[j].is_zero()) {
				S.add(TA.vr[j]);
			}
			else
			{
				V.vr.push_back(T.vr[j]);
				V.vr_i.push_back(T.vr_i[j]);
			}
		RV=V;
		RS=S;
		
		return dim_S;
	};
	
	friend vs2T meet(const vs2T& A,  const vs2T& B)
	{ // A meet B
		vs2T V;
		vs2T S;
		
		span_and_meet(A,B,V,S);
		
		return S;
	};
};

#endif // _vs2T_h

